#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fts.h>
#include <regex.h>

#ifndef bool
#define bool int
#define true 1
#define false 0
#endif

#define DIR_SEP '/'
#define DIR_SEP_STR "/"

#define PROGRAM_NAME "path"
#define PROGRAM_VERSION "1.0"
#define PROGRAM_DESCRIPTION "Personal `locate' program."

#define DB_BASENAME "PATH"
#define DB_TAG "PATHDB"
#define HEADER_SIZE (sizeof(db_header) - sizeof(char *))

#define CONFIG_SIZE_MAX 10240

typedef enum error_type {
    FATAL_ERROR,
    ERROR,
    WARNING,
} error_type;

typedef enum code_type {
    CODE_NONE,
    CODE_ED, /* enter directory */
    CODE_LD, /* leave directory */
    CODE_LAST,
} code_type;

typedef struct db_header {
    char tag[6]; /* PATHDB */
    unsigned short config_size;
    char *config;
} db_header;

typedef struct string_list {
    char *str;
    struct string_list *next;
} string_list;

typedef struct regex_list {
    regex_t *regex;
    char *regex_str;
    struct regex_list *next;
} regex_list;

typedef void (*db_callback)(const db_header *header,
                            const char *name, short namelen,
                            const char *path, short pathlen);

/* prototypes */

void error(error_type type, const char *format, ...);

/* global variables */

char *program_name;
char *rootdir;
char *db;
char *search_str;
regex_t *search_regex;
bool search_match_regex;
string_list *excludes;
regex_list *exclude_regexes;
bool match_path;
bool clean_update;
char *ext_arg;
string_list *names;
bool quote;

/* util functions */

void error(error_type type, const char *format, ...) {
    int n;
    char *str;
    size_t size = 1024;
    va_list ap;

    if ((str = malloc(size)) == NULL)
        fprintf(stderr, "fatal error: error");

    while (true) {
        va_start(ap, format);
        n = vsnprintf(str, size, format, ap);
        va_end(ap);

        if (n > -1 && n < size) {
            switch (type) {
            case FATAL_ERROR:
                fprintf(stderr, "fatal error: %s", str);
                exit(127);
                break;
            case WARNING:
                fprintf(stderr, "warning: %s", str);
                break;
            default:
                fprintf(stderr, "%s", str);
                exit(1);
                break;
            }
            break;
        }

        size *= 2;
        if ((str = realloc(str, size)) == NULL)
            fprintf(stderr, "fatal error: error");
    }
}

string_list *string_list_new(const char *str) {
    string_list *list;

    if ((list = (string_list *)malloc(sizeof(string_list))) == NULL)
        error(FATAL_ERROR, "string_list_new: malloc: %s\n", strerror(errno));
    if ((list->str = strdup(str)) == NULL)
        error(FATAL_ERROR, "string_list_new: strdup: %s\n", strerror(errno));
    list->next = NULL;

    return list;
}

void string_list_delete(string_list *list) {
    string_list *tmp;

    while (list) {
        tmp = list->next;
        free(list->str);
        free(list);
        list = tmp;
    }
}

void string_list_push(string_list **list, string_list *elem) {
    if (!*list)
        *list = elem;
    else {
        string_list *p = *list;
        while (p->next)
            p = p->next;
        p->next = elem;
    }
}

static int string_compar(const void *p1, const void *p2) {
    return strcmp(*(char * const *)p1, *(char * const *)p2);
}

void sort_strings(char **array, size_t size) {
    qsort(array, size, sizeof(char *), string_compar);
}

regex_list *regex_list_new(const char *str) {
    regex_list *list;

    if ((list = (regex_list *)malloc(sizeof(regex_list))) == NULL)
        error(FATAL_ERROR, "regex_list_new: malloc: %s\n", strerror(errno));
    if ((list->regex = (regex_t *)malloc(sizeof(regex_t))) == NULL)
        error(FATAL_ERROR, "regex_list_new: malloc: %s\n", strerror(errno));
    if ((list->regex_str = strdup(str)) == NULL)
        error(FATAL_ERROR, "string_list_new: strdup: %s\n", strerror(errno));
    if (regcomp(list->regex, str, REG_EXTENDED) != 0)
        error(ERROR, "invalid exclued regex %s\n", str);
    list->next = NULL;
    
    return list;
}

void regex_list_delete(regex_list *list) {
    regex_list *tmp;

    while (list) {
        tmp = list->next;
        regfree(list->regex);
        free(list->regex);
        free(list->regex_str);
        free(list);
        list = tmp;
    }
}

void regex_list_push(regex_list **list, regex_list *elem) {
    if (!*list)
        *list = elem;
    else {
        regex_list *p = *list;
        while (p->next)
            p = p->next;
        p->next = elem;
    }
}

void path_concat(char *a, const char *b) {
    size_t a_len = strlen(a);
    if (a_len == 0)
        strcpy(a, b);
    else {
        char *p = a + a_len - 1;
        if (*b == DIR_SEP) {
            if (*p == DIR_SEP)
                *p = '\0';
        } else {
            if (*p != DIR_SEP) {
                *(p + 1) = DIR_SEP;
                *(p + 2) = '\0';
            }
        }
        strcat(a, b);
    }
}

void path_concat_fast(char *a, size_t a_len, const char *b, size_t b_len) {
    a[a_len] = DIR_SEP;
    memcpy(a + a_len + 1, b, b_len);
    a[a_len + b_len + 1] = '\0';
}

char *path_alloc_concat(const char *a, const char *b) {
    char *ptr;
    size_t a_len = strlen(a);
    size_t b_len = strlen(b);
    size_t ab_len = a_len + b_len + 2;
    
    if (ab_len < a_len || ab_len < b_len)
        error(FATAL_ERROR, "path_concat_dup: integer overflow\n");
    
    if ((ptr = malloc(ab_len)) == NULL)
        error(FATAL_ERROR, "path_alloc_concat: malloc: %s\n", strerror(errno));
    
    strcpy(ptr, a);
    path_concat(ptr, b);
    
    return ptr;
}

char *get_cwd(size_t extra) {
    char *buf = NULL;
    
    size_t size = 1024;
    
    if ((buf = (char *)malloc(size + extra)) == NULL)
        error(FATAL_ERROR, "get_cwd: malloc: %s\n", strerror(errno));
    
    while (true) {
        if (getcwd(buf, size) == NULL) {
            if (errno == ERANGE) {
                size *= 2;
                if ((buf = (char *)realloc(buf, size + extra)) == NULL)
                    error(FATAL_ERROR, "get_cwd: malloc: %s\n", strerror(errno));
            } else
                error(FATAL_ERROR, "get_cwd: getcwd: %s\n", strerror(errno));
        } else
            break;
    }

    if (!buf)
        error(FATAL_ERROR, "get_cwd: getcwd\n");

    return buf;
}

char *get_absolute_path(char *path) {
    char *buf = NULL;
    
    if (*path == DIR_SEP) {
        if ((buf = strdup(path)) == NULL)
            error(FATAL_ERROR, "get_absolute_path: strdup: %s\n", strerror(errno));
    } else {
        size_t len = strlen(path);
        buf = get_cwd(len + 2);
        path_concat_fast(buf, strlen(buf), path, len);
    }
    
    return buf;
}

/* core functions */

void init(int argc, char *argv[]) {
    if ((program_name = strdup(argc > 0 ? argv[0] : PROGRAM_NAME)) == NULL)
        error(FATAL_ERROR, "init: malloc: %s\n", strerror(errno));
}

void cleanup() {
    free(program_name);
    free(rootdir);
    free(db);
    free(search_str);
    if (search_regex) {
        regfree(search_regex);
        free(search_regex);
    }
    string_list_delete(excludes);
    regex_list_delete(exclude_regexes);
    free(ext_arg);
    string_list_delete(names);
}

void add_exclude_str(const char *str) {
    string_list_push(&excludes, string_list_new(str));
}

void add_exclude_regex(const char *regex) {
    regex_list_push(&exclude_regexes, regex_list_new(regex));
}

bool match_exclude_entry(FTSENT *ftsent) {
    const char *str = !match_path ? ftsent->fts_name : ftsent->fts_path;

    if (excludes) {
        string_list *p = excludes;
        while (p) {
            if (strstr(str, p->str) != NULL)
                return true;
            p = p->next;
        }
    }

    if (exclude_regexes) {
        regex_list *p = exclude_regexes;
        while (p) {
            if (regexec(p->regex, str, 0, NULL, 0) == 0)
                return true;
            p = p->next;
        }
    }

    return false;
}

char *serialize_config(unsigned short *psize) {
    if (excludes || exclude_regexes) {
        char *config;
        unsigned short index = 0;
        unsigned short size = 0;

        /* calculate size of config */

        if (excludes) {
            string_list *p = excludes;
            while (p) {
                size += 1 + strlen(p->str) + 1;
                p = p->next;
            }
        }
        if (exclude_regexes) {
            regex_list *p = exclude_regexes;
            while (p) {
                size += 1 + strlen(p->regex_str) + 1;
                p = p->next;
            }
        }
        size++;
        *psize = size;

        if (size > CONFIG_SIZE_MAX)
            error(FATAL_ERROR, "reached max config size\n", CONFIG_SIZE_MAX);

        /* do serialize */

        if ((config = (char *)malloc(size + 1)) == NULL)
            error(FATAL_ERROR, "serialize_config: malloc: %s\n", strerror(errno));
        
        if (excludes) {
            string_list *p = excludes;
            size_t len;
            while (p) {
                len = strlen(p->str);
                config[index++] = 'e';
                memcpy(config + index, p->str, len + 1);
                index += len + 1;
                p = p->next;
            }
        }

        if (excludes) {
            regex_list *p = exclude_regexes;
            size_t len;
            while (p) {
                len = strlen(p->regex_str);
                config[index++] = 'E';
                memcpy(config + index, p->regex_str, len + 1);
                index += len + 1;
                p = p->next;
            }
        }

        config[index++] = '\0';

        return config;
    } else
        return NULL;
}

void unserialize_config(const char *config, unsigned short size) {
    const char *p = config;
    char opt, c;
    char *arg;
    int index;
    int arg_size = 1024;

    if (size > CONFIG_SIZE_MAX)
        error(FATAL_ERROR, "reached max config size\n", CONFIG_SIZE_MAX);

    if ((arg = (char *)malloc(arg_size)) == NULL)
        error(FATAL_ERROR, "unserialize_config: malloc: %s\n", strerror(errno));

    while (size > 0 && *p != '\0') {
        opt = *p++;
        size--;

        if (opt != 'e' && opt != 'E')
            error(FATAL_ERROR, "invalid config option %c\n", opt);
        
        /* parse arg */
        index = 0;
        while (size > 0) {
            size--;
            
            if (index == arg_size - 1) {
                arg_size *= 2;
                if (arg_size < 0)
                    error(FATAL_ERROR, "unserialize_config: interger overflow\n");
                if ((arg = (char *)realloc(arg, arg_size)) == NULL)
                    error(FATAL_ERROR, "unserialize_config: realloc: %s\n", strerror(errno));
            }
            
            arg[index++] = *p++;

            if (size > 0 && *p == '\0') {
                p++;
                break;
            }
        }
        arg[index] = '\0';

        switch (opt) {
        case 'e':
            add_exclude_str(arg);
            break;
        case 'E':
            add_exclude_regex(arg);
            break;
        }
    }

    if (*p != '\0')
        error(FATAL_ERROR, "invalid config\n");

    free(arg);
}

db_header *new_db_header() {
    db_header *header = (db_header *)malloc(HEADER_SIZE);
    if (!header)
        error(FATAL_ERROR, "new_db_header: malloc: %s\n", strerror(errno));
    memset(header, 0, sizeof(db_header));
    return header;
}

void delete_db_header(db_header *header) {
    free(header->config);
    free(header);
}

void read_db_header(FILE *fp, db_header *header) {
    char * config = NULL;
    unsigned short config_size;

    if (fread(header, HEADER_SIZE, 1, fp) != 1)
        error(ERROR, "cannot read db header: %s\n", strerror(errno));

    if (memcmp(header->tag, DB_TAG, strlen(DB_TAG)) != 0)
        error(ERROR, "cannot identify tag\n");

    if (header->config_size > CONFIG_SIZE_MAX)
        error(FATAL_ERROR, "reached max config size\n", CONFIG_SIZE_MAX);
    config_size = header->config_size;

    if (config_size > 0) {
        if ((config = (char *)malloc(config_size)) == NULL)
            error(FATAL_ERROR, "read_db_header: malloc: %s\n", strerror(errno));

        if (fread(config, config_size, 1, fp) != 1)
            error(FATAL_ERROR, "read_db_header: fread: %s\n", strerror(errno));
    }

    header->config = config;
}

void write_db_header(FILE *fp, const db_header *header) {
    if (fwrite(header, HEADER_SIZE, 1, fp) != 1)
        error(FATAL_ERROR, "write_db_header: fwrite: %s\n", strerror(errno));
    if (header->config && header->config_size) {
        if (fwrite(header->config, header->config_size, 1, fp) != 1)
            error(FATAL_ERROR, "write_db_header: fwrite: %s\n", strerror(errno));
    }
}

void read_db_body(FILE *fp, const db_header *header, db_callback callback) {
    char c, d;
    char *name;
    int namesize = 1024;
    int namelen = 0;
    char *path;
    int pathsize = 1024;
    int pathlen = 0;
    int *stack;
    int stacksize = 1024;
    int stackidx = -1;
    
    if ((name = (char *)malloc(namesize)) == NULL)
        error(FATAL_ERROR, "read_db_body: malloc: %s\n", strerror(errno));
    if ((path = (char *)malloc(pathsize)) == NULL)
        error(FATAL_ERROR, "read_db_body: malloc: %s\n", strerror(errno));
    if ((stack = (int *)malloc(sizeof(int) * stacksize)) == NULL)
        error(FATAL_ERROR, "read_db_body: malloc: %s\n", strerror(errno));

    path[0] = '\0';

    while (true) {
        c = fgetc(fp);
        if (c == EOF)
            break;

        if ((unsigned char)c < CODE_LAST) {
            if (c == CODE_ED) {
                if (stackidx == stacksize) {
                    stacksize *= 2;
                    if (stacksize < 0)
                        error(FATAL_ERROR, "read_db_body: integer overflow\n");
                    if ((stack = (int *)realloc(stack, sizeof(int) * stacksize)) == NULL)
                        error(FATAL_ERROR, "read_db_body: realloc: %s\n", strerror(errno));
                }
                stack[++stackidx] = pathlen;
            } else if (c == CODE_LD) {
                if (stackidx < 0)
                    error(FATAL_ERROR, "read_db_body: invalid input\n");
                pathlen = stack[stackidx--];
                *(path + pathlen) = '\0';
            } else
                error(FATAL_ERROR, "read_db_body: unknown code: %d\n", c);
        } else {
            namelen = 0;
            do {
                name[namelen++] = (char)c;
                c = fgetc(fp);
            } while (c != '\0' && c != EOF);
            name[namelen] = '\0';

            if (pathlen + namelen + 1 < 0)
                error(FATAL_ERROR, "read_db_body: interger overflow\n");
            while (pathlen + namelen + 1 > pathsize) {
                pathsize *= 2;
                if (pathsize < 0)
                    error(FATAL_ERROR, "read_db_body: integer overflow\n");
            }
            if ((path = (char *)realloc(path, pathsize)) == NULL)
                error(FATAL_ERROR, "read_db_body: malloc: %s\n", strerror(errno));

            path_concat_fast(path, pathlen, name, namelen);
            if (d == CODE_ED)
                pathlen += namelen + 1;
            
            /* TODO refactoring */
            callback(header, name, namelen, path + 1, pathlen - 1); /* trim slash */
        }

        d = c;
    }

    free(name);
    free(path);
    free(stack);
}

bool write_entry(FILE *fp, FTSENT *ftsent) {
    code_type code = CODE_NONE;
    bool done = false;


    /* barrier */
    if (ftsent->fts_info == FTS_NS)
        return false;
    else if (ftsent->fts_namelen == 0)
        return true;

    if (ftsent->fts_info == FTS_D)
        code = CODE_ED;
    else if (ftsent->fts_info == FTS_DP) {
        code = CODE_LD;
        done = true;
    }

    if (code != CODE_NONE) {
        if (fputc((int)code, fp) == EOF)
            error(FATAL_ERROR, "write_entry: fputc: %s\n", strerror(errno));
    }

    if (done)
        return true;

    if (fwrite(ftsent->fts_name, ftsent->fts_namelen, 1, fp) != 1)
        error(FATAL_ERROR, "write_entry: fwrite: %s\n", strerror(errno));
    if (fputc('\0', fp) == EOF)
        error(FATAL_ERROR, "write_entry: fputc: %s\n", strerror(errno));
    
    return true;
}

char *find_db() {
    char *buf, *p;
    int workspace = 16; /* magic number */

    if (rootdir) {
        size_t len = strlen(rootdir);
        if ((buf = (char *)malloc(len + workspace + 1)) == NULL)
            error(FATAL_ERROR, "find_db: malloc: %s\n", strerror(errno));
        memcpy(buf, rootdir, len + 1);
    } else
        buf = get_cwd(workspace);

    while (true) {
        if (access(buf, R_OK | X_OK) != 0)
            break;

        p = buf + strlen(buf);
        path_concat(buf, DB_BASENAME);

        if (access(buf, R_OK) == 0) {
            if (!rootdir) {
                if ((rootdir = strdup(buf)) == NULL)
                    error(FATAL_ERROR, "find_db: strdup: %s\n", strerror(errno));
                dirname(rootdir);
            }
            
            return buf;
        } else if (errno != ENOENT)
            break;

        *p = '\0';
        dirname(buf);

        if (buf[1] == '\0')
            break;
    }

    free(buf);
    
    return NULL;
}

void alter_db(db_callback callback) {
    db = find_db();
    
    if (db) {
        FILE *fp;
        db_header *header;

        fp = fopen(db, "rb");
        if (!fp)
            error(ERROR, "cannot open db %s: %s\n", db, strerror(errno));

        header = new_db_header();

        read_db_header(fp, header);
        read_db_body(fp, header, callback);

        delete_db_header(header);

        fclose(fp);
    } else
        error(ERROR, "cannot find db\n");
}

void update_db() {
    FILE *fp = NULL;
    FTS *fts = NULL;
    FTSENT *ftsent = NULL;
    char **dir = NULL;
    db_header *header = NULL;

    if (access(db, R_OK | W_OK) != 0) {
        if (errno == ENOENT) {
            fp = fopen(db, "wb");
            if (!fp)
                error(ERROR, "cannot open db %s: %s\n", db, strerror(errno));
        } else
            error(ERROR, "cannot access db %s: %s\n", db, strerror(errno));
    }

    if (!fp) {
        /* retrieve config from existed db */
        fp = fopen(db, "rb");
        if (!fp)
            error(FATAL_ERROR, "update_db: fopen %s: %s\n", db, strerror(errno));

        header = new_db_header();
        read_db_header(fp, header);

        if (!clean_update && header->config && header->config_size)
            unserialize_config(header->config, header->config_size);

        /* close current file and reopen it as overwrite */
        fclose(fp);
        fp = fopen(db, "wb");
        if (!fp)
            error(FATAL_ERROR, "update_db: fopen %s: %s\n", db, strerror(errno));
    }

    /* create db header if it is not existed */
    if (!header) {
        header = new_db_header();
        memcpy(header->tag, DB_TAG, sizeof(header->tag));

        header->config = serialize_config(&header->config_size);
    } else {
        if (!clean_update) {
            char *config;
            
            /* merge config */
            free(header->config);
            header->config_size = 0;
            header->config = serialize_config(&header->config_size);
        }
    }

    write_db_header(fp, header);

    dir = malloc(sizeof(char **) * 2);
    if (!dir)
        error(FATAL_ERROR, "update_db: malloc: %s\n", strerror(errno));
    *dir = path_alloc_concat(rootdir, "");
    dir[1] = NULL;
    
    fts = fts_open(dir, FTS_PHYSICAL, NULL);
    if (!fts)
        error(FATAL_ERROR, "update_db: fts_open: %s\n", strerror(errno));

    while (true) {
        ftsent = fts_read(fts);

        if (!ftsent)
            break;

        if (match_exclude_entry(ftsent) || !write_entry(fp, ftsent))
            fts_set(fts, ftsent, FTS_SKIP);
    }

    free(*dir);
    free(dir);

    fts_close(fts);

    fclose(fp);

    delete_db_header(header);
}

void search_callback(const db_header *header,
                    const char *name, short namelen,
                    const char *path, short pathlen) {
    bool matched = false;
    if (!search_match_regex) {
        if (strstr(match_path ? path : name, search_str) != NULL)
            matched = true;
    } else {
        if (regexec(search_regex, match_path ? path : name,
                    0, NULL, 0) == 0)
            matched = true;
    }

    if (matched) {
        printf(quote
               ? "\"%s%c%s\"\n"
               : "%s%c%s\n",
               rootdir, DIR_SEP, path);
    }
}

void list_names_callback(const db_header *header,
                         const char *name, short namelen,
                         const char *path, short pathlen) {
    string_list_push(&names, string_list_new(name));
}

/* commands */

void usage_command() {
    printf("Usage: %s\n\n",
           PROGRAM_NAME);

    printf(" Search string:\n"
           "  %s [-prQ] <search string>\n"
           " Update database:\n"
           "  %s -u [-p] [-d <database directory>] [-e <exclude string>]... [-E <exclude regex>]...\n"
           " Display database path:\n"
           "  %s -P\n"
           "\n",
           program_name, program_name, program_name, program_name);

    printf("Examples:\n"
           "  %s -u           # create/update a database for the current directory.\n"
           "  %s -u -d ~/     # create/update a database for ~/.\n"
           "  %s README       # list README locations from the current directory database.\n"
           "  %s -d ~/ README # list README locatiosn from ~/ database.\n"
           "  %s -e svn -u    # add exclusion `svn' into the current directory database and update it.\n"
           "  %s -U           # create a brand-new database for the current directory.\n"
           "\n",
           program_name, program_name, program_name, program_name,
           program_name, program_name);

    printf("Options:\n"
           "  -p                      - Match the path instead of the basename.\n"
           "  -P                      - Display the database path.\n"
           "  -r                      - Match using POSIX Extended Regular Expression.\n"
           "  -d <database directory> - Specifies a database directory.\n"
           "  -Q                      - Quote file locations.\n"
           "  -u                      - Update the database.\n"
           "  -U                      - Clean and update the database.\n"
           "  -e <exclude string>     - Exclude entries from the database if matched.\n"
           "                            This option is automatically applies next time until using -U.\n"
           "  -E <exclude regex>      - Exclude entries from the database using\n"
           "                            POSIX Extended Regular Expression if matched.\n"
           "                            This option is automatically applies next time until using -U.\n"
           "  -h                      - Display this message.\n"
           "  -V                      - Display version and copyright.\n");
}

void version_command() {
    printf("%s %s\n", PROGRAM_NAME, PROGRAM_VERSION);
    printf("Copyright (C) 2007 MATSUYAMA Tomohiro.\n"
           "This is free software. You may redistribute copies of it under the terms of\n"
           "the GNU General Public License version 2 or later.\n"
           "There is NO WARRANTY.\n");
}

void print_db_path() {
    db = find_db();

    if (db)
        printf("%s\n", db);
    else
        error(ERROR, "%s not found.\n", DB_BASENAME);
}

void update_db_command() {
    if (!rootdir)
        rootdir = get_cwd(0);

    db = path_alloc_concat(rootdir, DB_BASENAME);
    
    update_db();
}

void search_command() {
    alter_db(search_callback);
}

void ext_command() {
    if (strcmp(ext_arg, "list-names") == 0) {
        /* collect names */
        alter_db(list_names_callback);

        if (names) {
            char **array;
            size_t index;
            size_t size = 0;
            string_list *p;
            char *str;
            size_t len;

            /* calculate list size */
            p = names;
            while (p) {
                if (size + 1 < size)
                    error(FATAL_ERROR, "ext_command: integer overflow\n");
                size++;
                p = p->next;
            }

            /* convert list into array */
            if ((array = (char **)malloc(sizeof(char *) * size)) == NULL)
                error(FATAL_ERROR, "ext_command: malloc: %s\n", strerror(errno));

            index = 0;
            p = names;
            while (p) {
                array[index++] = p->str;
                p = p->next;
            }

            /* sort */
            sort_strings(array, size);

            str = array[0];
            len = strlen(str);
            for (index = 0; index < size; ) {
                index++;
                
                /* print */
                printf(quote
                       ? "\"%s\"\n"
                       : "%s\n",
                       str);

                /* skip same names */
                while (index < size) {
                    char *next_str = array[index];
                    size_t next_len = strlen(next_str);
                    if (len != next_len || memcmp(str, next_str, len) != 0) {
                        str = next_str;
                        len = next_len;
                        break;
                    } else
                        index++;
                }
            }

            free(array);
        }
    }
}

/* main */

int main(int argc, char *argv[]) {
    int opt;
    void (*command)() = usage_command;

    init(argc, argv);
    
    while ((opt = getopt(argc, argv, "pPrd:QuUe:E:x:hV")) != -1) {
        switch (opt) {
        case 'p':
            match_path = true;
            break;
        case 'P':
            command = print_db_path;
            break;
        case 'r':
            search_match_regex = true;
            break;
        case 'u':
            command = update_db_command;
            break;
        case 'U':
            command = update_db_command;
            clean_update = true;
            break;
        case 'd':
            rootdir = get_absolute_path(optarg);
            if (access(rootdir, R_OK | X_OK) != 0)
                error(ERROR, "cannot access %s: %s\n", rootdir, strerror(errno));
            else {
                size_t len = strlen(rootdir);
                if (len > 0) { 
                    while (len > 2 && rootdir[--len] == DIR_SEP)
                        rootdir[len] = '\0';
                }
            }
            break;
        case 'Q':
            quote = true;
            break;
        case 'e':
            add_exclude_str(optarg);
            break;
        case 'E':
            add_exclude_regex(optarg);
            break;
        case 'x':
            if ((ext_arg = strdup(optarg)) == NULL)
                error(FATAL_ERROR, "main: strdup: %s\n", strerror(errno));
            command = ext_command;
            break;
        case 'h':
            command = usage_command;
            break;
        case 'V':
            command = version_command;
            break;
        }
    }

    if (optind < argc && command == usage_command) {
        if (!search_match_regex) {
            if ((search_str = strdup(argv[optind])) == NULL)
                error(FATAL_ERROR, "main: strdup: %s\n", strerror(errno));
        } else {
            if ((search_regex = (regex_t *)malloc(sizeof(regex_t))) == NULL)
                error(FATAL_ERROR, "main: malloc: %s\n", strerror(errno));
            if (regcomp(search_regex, argv[optind], REG_EXTENDED) != 0)
                error(ERROR, "invalid regex %s\n", argv[optind]);
        }
        command = search_command;
    }

    command();

    cleanup();

    return 0;
}
