;;; path.el

;; Author: MATSUYAMA Tomohiro
;; Version: 1.0
;; Licence: GPLv2

;;; Usage
;; Add the following code into your .emacs.
;; (autoload 'path-goto-path "path" nil t)
;; (autoload 'path-find-file "path" nil t)
;; (global-set-key "\C-x\C-g" 'path-goto-path)
;; (global-set-key "\C-x\M-f" 'path-find-file)

;;; Code

(defconst path-db-name "PATH") 
(defconst path-buffer-name "*PATH*")

(defvar path-do-cache t)
(defvar path-hashtab (make-hash-table :test 'equal))

(defvar path-db-rootdir nil)
(defvar path-saved-buffer nil)

(defvar path-select-mode-map (make-sparse-keymap))
(define-key path-select-mode-map "n" 'next-line)
(define-key path-select-mode-map "p" 'previous-line)
(define-key path-select-mode-map "q" 'path-quit-selection)
(define-key path-select-mode-map "\C-m" 'path-select-path)

(defun path-get-buffer ()
  (let ((buffer (get-buffer path-buffer-name)))
    (when (null buffer)
      (setq buffer (generate-new-buffer (generate-new-buffer-name path-buffer-name)))
      (buffer-disable-undo buffer))
    buffer))

(defun path-call-process (&rest args)
  (let ((buffer (path-get-buffer))
        (saved-buffer (current-buffer))
        result)
    (set-buffer buffer)
    (when path-db-rootdir
        (cd path-db-rootdir))
    (delete-region (point-min) (point-max))
    (setq result
          (if (= 0 (apply 'call-process "path" nil t nil args))
              (buffer-substring (point-min) (1- (point-max)))
            nil))
    (set-buffer saved-buffer)
    result))

(defun path-find-db-rootdir (dir)
  (let (rootdir)
    (setq rootdir (path-call-process "-d" dir "-P"))
    (if (null rootdir)
        nil
      (file-name-directory rootdir))))

(defun path-list-names ()
  (path-visit-db-rootdir)
  (let ((list (gethash path-db-rootdir path-hashtab)))
    (when (null list)
      (setq list (split-string (path-call-process "-x" "list-names") "\n"))
      (when path-do-cache
        (puthash path-db-rootdir list path-hashtab)))
    list))

(defun path-clear-cache ()
  (interactive)
  (clrhash path-hashtab))

(defun path-visit-db-rootdir ()
  (interactive)
  (setq path-db-rootdir (path-find-db-rootdir (expand-file-name default-directory))))

(defun path-select-mode ()
  (interactive)
  (use-local-map path-select-mode-map)
  (setq major-mode 'path-select-mode
        mode-name "Path-Select")
  (goto-char (point-min))
  (run-hooks 'path-select-mode-hook))

(defun path-quit-selection ()
  (interactive)
  (switch-to-buffer path-saved-buffer))

(defun path-select-path ()
  (interactive)
  (find-file (buffer-substring (point-at-bol) (point-at-eol))))

;;;###autoload
(defun path-goto-path (name)
  (interactive "sGoto path: ")
  (path-visit-db-rootdir)
  (let ((str (path-call-process "-r" (format "^%s$" name))))
    (when str
      (let* ((list (split-string str "\n"))
             (count (safe-length list)))
        (cond
         ((= count 1)
          (find-file (car list)))
         ((> count 1)
          (setq path-saved-buffer (current-buffer))
          (switch-to-buffer (path-get-buffer))
          (path-select-mode)))))))

;;;###autoload
(defun path-find-file ()
  (interactive)
  (setq input (completing-read "Find file: " (path-list-names)))
  (when input
    (path-goto-path input)))
